/* This file is part of Cubix, a simple Basic interpreter
 * Copyright (C) 1999-2000 Ariya Hidayat, ariya@tf.itb.ac.id
 *
 * Cubix is free software; you can redistribute it and/or modify it
 * under the terms of the Artistic License (see the file LICENSE)
 */

%{

#include "cubix.h"
#include "gram.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* in cxengine.c */
struct CXSymbol* create_symbol (char*, int);
double create_number (char*);
char* create_string (char*);
void compile_error (long, char*, char*);

/* local stuff */
long compile_lineno;
int yywrap(void);
int yyerror(char*);
int is_keyword (char*);

%}

digit [0-9]
alphanum [a-zA-Z][a-zA-Z0-9]*
whitespace [ \t\f\r\v]
remark [Rr][Ee][Mm]

%x RET

%%

<<EOF>> { return ENDOFFILE;}

{whitespace}+ {BEGIN(INITIAL);}     /* ignore whitespace */

\n { ++compile_lineno; return SEPARATOR; }
: { return SEPARATOR; }

{remark}{whitespace}+.* { return SEPARATOR;}  /* comments span til eol */
#+.* { return SEPARATOR;}  /* UNIX-style comments */
{remark}\n { compile_lineno++; return SEPARATOR;}
{remark} {yymore();}


{alphanum}+  {
   int i = is_keyword(yytext);
   if (i) return i;
   yylval.symbol = create_symbol(yytext, CUBIX_NUMBER);
   return SYMBOL;
   }

{alphanum}+"$" {
   int i = is_keyword(yytext);
   if (i) return i;
   yylval.symbol = create_symbol(yytext, CUBIX_STRING);
   return STRSYMBOL;
}

(({digit}+|({digit}*\.{digit}*))([eE][-+]?{digit}+)?) {
   yylval.number = create_number(yytext);
   return NUMBER;
}

\"([^\"\n]|"\\\"")*\" {
  yylval.string = create_string(yytext);
  return STRING;
}

"=" return SEQ;
"<>" return SNE;
"<" return SLT;
"<=" return SLE;
">" return SGT;
">=" return SGE;


. return *yytext;


%%

/* dummy function so we won't need libfl.a */
int yywrap(){ return 1; }

/* spawn to compile_error() in cxengine.c */
int yyerror (char *msg)
{
  extern char* yytext;
  compile_error(compile_lineno, msg, yytext);
  return 0;
}

/* return appropriate token if given string is keyword */
#include <string.h>
int is_keyword (char *s)
{
  int i;
  struct key { char* name; int token; }
  keywords[] = {
    { "external",   SEXTERNAL },
    { "sub",        SSUB     },
    { "subroutine", SSUB     },
    { "end",        SEND     },
    { "not",        SNOT     },
    { "or",         SOR      },
    { "and",        SAND     },
    { "xor",        SXOR     },
    { "print",      SPRINT   },
    { "if",         SIF      },
    { "then",       STHEN    },
    { "else",       SELSE    },
    { "endif",      SENDIF   },
    { "for",        SFOR     },
    { "to",         STO      },
    { "step",       SSTEP    },
    { "next",       SNEXT    },
    { "sin",        MSIN     },
    { "asin",       MASIN    },
    { "cos",        MCOS     },
    { "acos",       MACOS    },
    { "tan",        MTAN     },
    { "atan",       MATAN    },
    { "exp",        MEXP     },
    { "log",        MLOG     },
    { "sqrt",       MSQRT    },
    { "int",        MINT     },
    { "frac",       MFRAC    },
    { "ran",        MRAN     },
    { "len",        MLEN     },
    { "val",        MVAL     },
    { "asc",        MASC     },
    { "left$",      SLEFT    },
    { "right$",     SRIGHT   },
    { "mid$",       SMID     },
    { "date$",      SDATE    },
    { "time$",      STIME    },
    { "str$",       SSTR     },
    { "chr$",       SCHR     },
    { "lower$",     SLOWER   },
    { "upper$",     SUPPER   },
    { (char*) 0L, 0 }};

  for (i = 0; keywords[i].token; i++)
    if(!strcasecmp(s, keywords[i].name)) return keywords[i].token;
  return 0;
}
