/* This file is part of Cubix, a simple Basic interpreter
 * Copyright (C) 1999-2000 Ariya Hidayat, ariya@tf.itb.ac.id
 *
 * Cubix is free software; you can redistribute it and/or modify it
 * under the terms of the Artistic License (see the file LICENSE)
 */

%{
#include "cubix.h"

struct CXSymbol* create_symbol (char*, int);
double convert_number (char*);
char* convert_string (char*, char*);

void externalize (struct CXSymbol*);
void begin_subroutine(char*);
void gen_pushsymbol (struct CXSymbol*);
void gen_popsymbol (struct CXSymbol*);
void gen_pushnumber (double);
void gen_pushstring (char*);
void gen_jump (char*);
void gen_label (char*);
void gen_if (void);
void gen_for (struct CXSymbol*, double step);
void gen_next (struct CXSymbol*, double);
void gen_then (void);
void gen_endif (void);
void gen_function (int);
void gen_call(char*);
void gen_return(void);

%}

%union {
  struct CXSymbol* symbol;
  double number;
  char* string;
  }

%token <symbol> SYMBOL
%token <symbol> STRSYMBOL
%token <number> NUMBER
%token <string> STRING

%token SEXTERNAL SSUB
%token SIF STHEN SELSE SENDIF SFOR STO SSTEP SNEXT
%token SPRINT SEND
%token SNOT SOR SAND SXOR SEQ SNE SLT SLE SGT SGE
%token MSIN MASIN MCOS MACOS MTAN MATAN
%token MEXP MLOG MSQRT MINT MFRAC MRAN
%token MLEN MVAL MASC

%token SLOWER SUPPER SLEFT SRIGHT SMID SDATE STIME SSTR SCHR

%token ENDOFFILE
%token SEPARATOR

%left SEPARATOR

%left SOR
%left SAND
%left SXOR

%left SEQ SNE SLT SLE SGT SGE

%left ','
%left '-' '+'
%left '*' '/'
%left '^'

%left '|' '&' '~'

%nonassoc UBITNOT
%nonassoc UNOT
%nonassoc UMINUS

%type <number> loopstep
%type <number> constant

%%

program:
  statement_list ENDOFFILE { YYACCEPT; }
;

statement_list:
  statement
| statement_list SEPARATOR statement
;

statement:  /* empty */
  | string_assignment
  | assignment
  | loop
  | declaration
  | subroutine
  | subroutine_call
  | conditional
  | SPRINT printlist semicolon
;

declaration:
  external_declaration
;

external_declaration:
  SEXTERNAL external_list
;

external_list:
    external_item
  | external_list ',' external_item
;

external_item:
  SYMBOL { externalize($1); }
| STRSYMBOL { externalize($1); }

assignment:
  SYMBOL SEQ expression { gen_popsymbol($1); }
;

string_assignment:
  STRSYMBOL SEQ string_expression { gen_popsymbol($1); }
;

string_expression:
  STRSYMBOL { gen_pushsymbol($1); }
| string_function
| STRING { gen_pushstring($1); }
| string_expression '+' string_expression { gen_function(STRING_CONCAT); }
;

string_function:
 | SLEFT '(' string_expression ',' expression ')' { gen_function(FUNC_LEFT); }
 | SRIGHT '(' string_expression ',' expression ')' { gen_function(FUNC_RIGHT); }
 | SMID '(' string_expression ',' expression ',' expression ')'
    { gen_function(FUNC_MID); }
 | SDATE  { gen_function(FUNC_DATE); }
 | STIME  { gen_function(FUNC_TIME); }
 | SSTR '(' expression ')' { gen_function(FUNC_STR); }
 | SCHR '(' expression ')' { gen_function(FUNC_CHR); }
 | SLOWER '(' string_expression ')' { gen_function(STRING_LOWER); }
 | SUPPER '(' string_expression ')' { gen_function(STRING_UPPER); }
;

expression:
  NUMBER { gen_pushnumber($1); }
| function
| subroutine_call
| SYMBOL { gen_pushsymbol($1); }
| '(' expression ')'
| expression '+' expression { gen_function(BINARY_ADD);}
| expression '-' expression { gen_function(BINARY_SUB);}
| expression '*' expression { gen_function(BINARY_MUL);}
| expression '/' expression { gen_function(BINARY_DIV);}
| expression '^' expression { gen_function(BINARY_POWER);}
| expression SOR expression { gen_function(BITWISE_OR); }
| expression SAND expression { gen_function(BITWISE_AND); }
| expression SXOR expression { gen_function(BITWISE_XOR); }
| '-' expression %prec UMINUS { gen_function(UNARY_NEGATE);}
| SNOT expression %prec UBITNOT { gen_function(BITWISE_NOT); }
;


function:
  MSIN '(' expression ')'  { gen_function(FUNC_SIN); }
| MASIN '(' expression ')' { gen_function(FUNC_ASIN); }
| MCOS '(' expression ')'  { gen_function(FUNC_COS); }
| MACOS '(' expression ')' { gen_function(FUNC_ACOS);}
| MTAN '(' expression ')'  { gen_function(FUNC_TAN); }
| MATAN '(' expression ')' { gen_function(FUNC_ATAN); }
| MEXP '(' expression ')'  { gen_function(FUNC_EXP); }
| MLOG '(' expression ')'  { gen_function(FUNC_LOG); }
| MSQRT '(' expression ')' { gen_function(FUNC_SQRT); }
| MINT '(' expression ')'  { gen_function(FUNC_INT); }
| MFRAC '(' expression ')' { gen_function(FUNC_FRAC); }
| MRAN '(' expression ')'  { gen_function(FUNC_RAN); }
| MLEN '(' string_expression ')' { gen_function(FUNC_LEN); }
| MVAL '(' string_expression ')' { gen_function(FUNC_VAL); }
| MASC '(' string_expression ')' { gen_function(FUNC_ASC); }
;

subroutine:
  SSUB SYMBOL { begin_subroutine($2->name); }
| SSUB STRSYMBOL { begin_subroutine($2->name); }
| SEND SSUB { gen_return(); }
;

subroutine_call:
  SYMBOL '(' call_list ')' { gen_call($1->name); }
;

call_list:
  call_item
| call_list ',' call_item
;

call_item:
  expression
| string_expression
;

conditional:
  SIF comparison { gen_if(); }
  STHEN statement_list { gen_then(); }
  conditional_else
  SENDIF { gen_endif(); }
  ;

conditional_else:
  SELSE statement_list
| /* empty, no else */
;

loop:
  SFOR SYMBOL SEQ expression STO expression loopstep { gen_for ($2, $7); }
  SEPARATOR statement_list SNEXT { gen_next($2, $7); }
;

loopstep:
  SSTEP constant { $$ = $2 }
| { $$ = 1.0 } /* empty, default to 1 */
;

constant:
 NUMBER {$$=$1} | '+' NUMBER {$$=$2} | '-' NUMBER {$$=-$2;};

printlist:
  expression { gen_function (FUNC_PRINT); }
| string_expression { gen_function (FUNC_PRINT); }
| printlist ',' expression
   { gen_function(FUNC_PRINTSPACE); gen_function(FUNC_PRINT); }
| printlist ',' string_expression
   { gen_function(FUNC_PRINTSPACE); gen_function(FUNC_PRINT); }
;

semicolon:
  ';'
| { gen_function(FUNC_PRINTLN); }
  ;

comparison:
|  '(' comparison ')'
| comparison SOR comparison { gen_function(BOOLE_OR); }
| comparison SAND comparison { gen_function(BOOLE_AND); }
| SNOT comparison %prec UNOT { gen_function(BOOLE_NOT); }
| expression SEQ expression { gen_function(COMPARE_EQ);  }
| expression SNE expression { gen_function(COMPARE_NE);  }
| expression SLT expression { gen_function(COMPARE_LT);  }
| expression SLE expression { gen_function(COMPARE_LE);  }
| expression SGT expression { gen_function(COMPARE_GT);  }
| expression SGE expression { gen_function(COMPARE_GE);  }
| string_expression SEQ string_expression { gen_function(STRING_EQ); }
| string_expression SNE string_expression { gen_function(STRING_NE); };

%%
