/* This file is part of Cubix, a simple Basic interpreter
 * Copyright (C) 1999-2000 Ariya Hidayat, ariya@tf.itb.ac.id
 *
 * Cubix is free software; you can redistribute it and/or modify it
 * under the terms of the Artistic License (see the file LICENSE)
 */

#ifndef __CUBIX_H
#define __CUBIX_H

/* Cubix error code */
#define CXERROR_OK 0
#define CXERROR_SOURCEFILE 1
#define CXERROR_PARSE 2
#define CXERROR_MEMORY 3
#define CXERROR_NOBYTECODE 4
#define CXERROR_INVALIDBYTECODE 5
#define CXERROR_NESTEDSUB 6

#define CXERROR_STACK 20
#define CXERROR_NULLREF 6
#define CXERROR_BADSYMTYPE 7
#define CXERROR_TYPEMISMATCH 8
#define CXERROR_DIVZERO 9
#define CXERROR_DOMERR 10
#define CXERROR_NEGSQRT 11
#define CXERROR_SUB 12

/* type of variable/symbol, simple isn't it */
#define CUBIX_NUMBER 1
#define CUBIX_STRING 2
#define CUBIX_ADDRESS 3

/* constants for Cubix context state */
#define CUBIX_INITIALIZED 1
#define CUBIX_COMPILE 2
#define CUBIX_EXECUTE 3
#define CUBIX_ERROR 4
#define CUBIX_TERMINATED 5

/* bytecode (kind) */
enum {
  PCODE_TERMINATE = 1,
  PCODE_PUSHSYMBOL, PCODE_POPSYMBOL, PCODE_PUSHNUMBER, PCODE_PUSHSTRING,
  PCODE_JUMP, PCODE_LABEL, PCODE_IF, PCODE_FUNCTION,
  PCODE_CALL, PCODE_RETURN
  };

/* bytecode (function) */
enum {
  SWAP_STACK = 1, COPY_STACK, POP_STACK,
  UNARY_NEGATE,
  BINARY_ADD, BINARY_SUB, BINARY_MUL, BINARY_DIV, BINARY_POWER,
  BITWISE_OR, BITWISE_NOT, BITWISE_AND, BITWISE_XOR,
  BOOLE_NOT, BOOLE_OR, BOOLE_AND,
  COMPARE_EQ, COMPARE_NE, COMPARE_LT, COMPARE_LE, COMPARE_GT, COMPARE_GE,
  STRING_LOWER, STRING_UPPER,
  STRING_CONCAT, STRING_EQ, STRING_NE,
  FUNC_PRINT, FUNC_PRINTSPACE,FUNC_PRINTLN, 
  FUNC_SIN, FUNC_ASIN, FUNC_COS, FUNC_ACOS, FUNC_TAN, FUNC_ATAN,
  FUNC_EXP, FUNC_LOG, FUNC_SQRT, FUNC_INT, FUNC_FRAC, FUNC_RAN,
  FUNC_LEN, FUNC_VAL, FUNC_ASC, FUNC_LEFT, FUNC_RIGHT, FUNC_MID,
  FUNC_TIME, FUNC_DATE,
  FUNC_STR, FUNC_CHR
  };

/* callback functions for Cubix context */
typedef void (*CUBIX_CALLBACK_ERROR)(char*);
typedef void (*CUBIX_CALLBACK_SETNUMBER)(char*, double);
typedef void (*CUBIX_CALLBACK_SETSTRING)(char*, char*);
typedef double (*CUBIX_CALLBACK_GETNUMBER)(char*);
typedef char* (*CUBIX_CALLBACK_GETSTRING)(char*);

/* Cubix symbol */
struct CXSymbol {
  char* name;                 /* name of this symbol */
  int type;                   /* CUBIX_NUMBER or CUBIX_STRING */
  int external;               /* true if external (defined somewhere else) */
  double number;              /* value if is CUBIX_NUMBER */
  char* string;               /* points to string if is CUBIX_STRING */
  struct CXSymbol* next;      /* next symbol, NULL if last in the list */
  };

/* Cubix bytecode */
struct CXBytecode {
  int kind;                    /* type of bytecode */
  int function;                /* if PCODE_FUNCTION, function code number */
  double number;               /* immediate value to push */
  struct CXSymbol* symbol;     /* related symbol to push or pop */
  char* label;                 /* soon to be deleted */
  char* string;                /* immediate string to push */
  struct CXBytecode* next;     /* next bytecode, NULL if last in the list */
  };

/* Cubix subroutine */
struct CXSubroutine {
  char* name;
  struct CXSymbol* symbols;
  struct CXBytecode* bytecodes;
  struct CXSubroutine* next;
  };

/* Cubix stack */
struct CXStack {
  int type;                    /* either CUBIX_NUMBER of CUBIX_STRING */
  double number;               /* value if type is CUBIX_NUMBER */
  char *string;                /* pointer to string if type is CUBIX_STRING */
  struct CXStack* next;        /* next entry, NULL for single entry */
  struct CXBytecode* code;     /* address (for use with RETURN) */
  };

/* Cubix context */
struct CXContext {
  int state;                                     /* state of the context */
  char *source;                                  /* source filename */
  long lineno;                                   /* active line number */
  int errno;                                     /* hold latest error code */
  struct CXBytecode* execpoint;                  /* current execution point */
  struct CXBytecode* bytecodes;                  /* list of compiled codes */
  struct CXStack* stacks;                        /* stack for this context */
  struct CXSymbol* symbols;                      /* main list of symbols */
  struct CXSubroutine* subroutines;              /* user-defined subroutines */
  struct CXSubroutine* activesub;                /* sub in execution */
  CUBIX_CALLBACK_ERROR error_function;           /* callback functions */
  CUBIX_CALLBACK_SETNUMBER setnumber_function;
  CUBIX_CALLBACK_SETSTRING setstring_function;
  CUBIX_CALLBACK_GETNUMBER getnumber_function;
  CUBIX_CALLBACK_GETSTRING getstring_function;
  };

/* Cubix engine :> */
void CXInitialize(struct CXContext*);
void CXCompile(struct CXContext*, char*);
void CXExecute(struct CXContext*);
void CXExecuteSub(struct CXContext*, char*);
void CXRelease(struct CXContext*);
char* CXDescribeError(struct CXContext*);

#endif