/* This file is part of Cubix, a simple Basic interpreter
 * Copyright (C) 1999-2000 Ariya Hidayat, ariya@tf.itb.ac.id
 *
 * Cubix is free software; you can redistribute it and/or modify it
 * under the terms of the Artistic License (see the file LICENSE)
 */

#include <ctype.h>
#include <math.h>
//#include <mem.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "cubix.h"

#ifndef MAX_ID_LENGTH
#define MAX_ID_LENGTH 255
#endif

/*
 * content of this file:
 * ----------------------
 *   prototypes for exported function not listed in cubix.h
 *   definition and declaration of local structs, variables, etc
 *   prototypes for memory management functions
 *   prototypes for local functions
 *   exported function not visible from cubix.h
 *   exported functions visible from cubix.h
 *   memory management functions
 *   local functions
*/

/* ========================================================================= */
/* EXPORTED FUNCTION PROTOTYPE */
/* ========================================================================= */

/* called from basic.l  */
struct CXSymbol* create_symbol(char*, int);
double create_number(char*);
char* create_string(char*);

/* called from basic.y */
void externalize(struct CXSymbol*);
void begin_subroutine(char*);
void gen_pushsymbol(struct CXSymbol*);
void gen_popsymbol(struct CXSymbol*);
void gen_pushnumber(double);
void gen_pushstring(char*);
void gen_if(void);
void gen_for(struct CXSymbol*, double step);
void gen_next(struct CXSymbol*, double);
void gen_then(void);
void gen_endif(void);
void gen_function(int);
void gen_call(char*);
void gen_return(void);


/* ========================================================================= */
/*  LOCAL STUFF (structure, variables, etc) */
/* ========================================================================= */

/* non-zero if compilation error occured */
int compile_error_code;

/* autolabel for loop and conditional */
int autolabel;

/* internal compiler stack */
#define MAX_INT_STACK 2048
int intstack_count;
int intstacks[MAX_INT_STACK];
char* magiclabel(int i){
static char name[MAX_ID_LENGTH];
sprintf (name,"magic_%d",i);
return name;
}

/* for the compiler, should be copied to the context after compilation */
struct CXSymbol* symbols;
struct CXBytecode* bytecodes;
struct CXSubroutine* subroutines;

/* point to current subroutine in progress, NULL if we're in main program */
struct CXSubroutine* current_sub;

/* ========================================================================= */
/*  MEMORY MANAGEMENT FUNCTION PROTOTYPE */
/* ========================================================================= */

int cubix_malloc_count = 0;
void* cubix_malloc(size_t);
void* cubix_free(void*);
char* cubix_strdup(char*);

/* ========================================================================= */
/*  LOCAL FUNCTION PROTOTYPE (for the compiler) */
/* ========================================================================= */

/* code generation */
struct CXBytecode* code_add(void);
void gen_jump(char*);
void gen_label(char*);
void gen_terminate(void);

/* deal with internal compiler stack */
void pushint(int);
int popint(void);
void swapint(void);

/* macro: add code with error-checking */
#define ADD_CODE(code) { \
  if (compile_error_code) return; \
  (code) = code_add(); \
  if (!(code)) return; \
  if (compile_error_code) return; \
  }

/* ========================================================================= */
/*  LOCAL FUNCTION PROTOTYPE (for the virtual-machine) */
/* ========================================================================= */

/* helper macro, these ain't a really good thing anyway */
#define MACHINE_EPSILON 1e-7
#define IS_SMALL(x) (fabs((x)) < MACHINE_EPSILON)
#define MAKE_LOGICAL(x)(int)((x) == 0.0) ? 0 : -1
#define CHECK_TYPE(cx,x,y) if((x)!=(y))trigger_error(cx, CXERROR_TYPEMISMATCH)
#define CHECK_ERROR(cx) if ((cx)->errno) return
#define STACK_PUSHNUMBER(cx,d){ \
   CHECK_ERROR((cx));\
   stack_pushnumber((cx), (d)); \
   CHECK_ERROR((cx));\
   };
#define STACK_PUSHSTRING(cx,s){ \
   CHECK_ERROR((cx));\
   stack_pushstring((cx), (s)); \
   CHECK_ERROR((cx)); \
   };
#define STACK_PUSHADDRESS(cx,a){ \
   CHECK_ERROR((cx));\
   stack_pushaddress((cx), (a)); \
   CHECK_ERROR((cx)); \
   };
#define STACK_POPNUMBER(cx,d){ \
   CHECK_ERROR((cx));\
   (d) = stack_popnumber((cx)); \
   CHECK_ERROR((cx));\
   };
#define STACK_POPSTRING(cx,s){ \
   CHECK_ERROR((cx));\
   (s) = stack_popstring((cx)); \
   CHECK_ERROR((cx));  \
   };

#define STACK_POPADDRESS(cx,a){ \
   CHECK_ERROR((cx));\
   (a) = stack_popaddress((cx)); \
   CHECK_ERROR((cx)); \
   };

/* code execution */
void single_step(struct CXContext*);
void trigger_error(struct CXContext*, int);
struct CXSubroutine* get_subroutine(struct CXContext*, char*);
struct CXStack* release_stacks(struct CXStack*);
struct CXSymbol* release_symbols(struct CXSymbol*);
struct CXBytecode* release_bytecodes(struct CXBytecode*);

/* stack handling */
void stack_copy (struct CXContext*);
void stack_swap (struct CXContext*);
void stack_pop (struct CXContext*);
struct CXStack* stack_pushnumber(struct CXContext*, double);
struct CXStack* stack_pushstring(struct CXContext*, char*);
struct CXStack* stack_pushaddress(struct CXContext*, struct CXBytecode*);
double stack_popnumber (struct CXContext* cx);
char* stack_popstring (struct CXContext* cx);
struct CXBytecode* stack_popaddress(struct CXContext*);
char* describe_function (int);

/* unconditional jump & call */
void program_jump(struct CXContext*, char*);
void program_call(struct CXContext*, char*);

/* run-time functions */
void function_negate(struct CXContext*);
void function_add(struct CXContext*);
void function_sub(struct CXContext*);
void function_mul(struct CXContext*);
void function_div(struct CXContext*);
void function_pow(struct CXContext*);
void function_not(struct CXContext*);
void function_or(struct CXContext*);
void function_and(struct CXContext*);
void function_xor(struct CXContext*);
void function_boole_not(struct CXContext*);
void function_boole_or(struct CXContext*);
void function_boole_and(struct CXContext*);
void function_compare_eq(struct CXContext*);
void function_compare_ne(struct CXContext*);
void function_compare_lt(struct CXContext*);
void function_compare_le(struct CXContext*);
void function_compare_gt(struct CXContext*);
void function_compare_ge(struct CXContext*);
void function_sin(struct CXContext*);
void function_asin(struct CXContext*);
void function_cos(struct CXContext*);
void function_acos(struct CXContext*);
void function_tan(struct CXContext*);
void function_atan(struct CXContext*);
void function_exp(struct CXContext*);
void function_log(struct CXContext*);
void function_sqrt(struct CXContext*);
void function_int(struct CXContext*);
void function_frac(struct CXContext*);
void function_ran(struct CXContext*);
void function_len(struct CXContext*);
void function_val(struct CXContext*);
void function_asc(struct CXContext*);
void function_left(struct CXContext*);
void function_right(struct CXContext*);
void function_mid(struct CXContext*);
void function_str(struct CXContext*);
void function_chr(struct CXContext*);
void function_date(struct CXContext*);
void function_time(struct CXContext*);
void function_string_lower(struct CXContext*);
void function_string_upper(struct CXContext*);
void function_string_concat(struct CXContext*);
void function_string_eq(struct CXContext*);
void function_string_ne(struct CXContext*);
void function_print(struct CXContext*);
void function_printspace(struct CXContext*);
void function_println(struct CXContext*);

/* functions look-up table */
struct function {
  int type;
  char* name;
  void (*proc)(struct CXContext*);
} functions[] = {
  { SWAP_STACK,   "SWAP",      stack_swap },
  { COPY_STACK,   "COPY",      stack_copy },
  { POP_STACK,    "POP",       stack_pop },
  { UNARY_NEGATE, "NEG",       function_negate },
  { BINARY_ADD,   "ADD",       function_add },
  { BINARY_SUB,   "SUB",       function_sub },
  { BINARY_MUL,   "MUL",       function_mul },
  { BINARY_DIV,   "DIV",       function_div },
  { BITWISE_NOT,  "NOT",       function_not },
  { BITWISE_OR,   "OR",        function_or },
  { BITWISE_AND,  "AND",       function_and },
  { BITWISE_XOR,  "XOR",       function_xor },
  { BOOLE_NOT,    "BOOLE NOT", function_boole_not },
  { BOOLE_OR,     "BOOLE OR",  function_boole_or },
  { BOOLE_AND,    "BOOLE AND", function_boole_and },
  { COMPARE_EQ,   "COMP EQ",   function_compare_eq },
  { COMPARE_NE,   "COMP NE",   function_compare_ne},
  { COMPARE_LT,   "COMP LT",   function_compare_lt },
  { COMPARE_LE,   "COMP LE",   function_compare_le },
  { COMPARE_GT,   "COMP GT",   function_compare_gt },
  { COMPARE_GE,   "COMP GE",   function_compare_ge },
  { FUNC_SIN,     "SIN",       function_sin },
  { FUNC_ASIN,    "ASIN",      function_asin },
  { FUNC_COS,     "COS",       function_cos },
  { FUNC_ACOS,    "ACOS",      function_acos },
  { FUNC_TAN,     "TAN",       function_tan },
  { FUNC_ATAN,    "ATAB",      function_atan },
  { FUNC_EXP,     "EXP",       function_exp },
  { FUNC_LOG,     "LOG",       function_log },
  { FUNC_SQRT,    "SQRT",      function_sqrt },
  { FUNC_INT,     "INT",       function_int },
  { FUNC_FRAC,    "FRAC",      function_frac },
  { FUNC_RAN,     "RAN",       function_ran },
  { FUNC_LEN,     "LEN",       function_len },
  { FUNC_VAL,     "VAL",       function_val },
  { FUNC_ASC,     "ASC",       function_asc },
  { FUNC_LEFT,    "LEFT",      function_left },
  { FUNC_RIGHT,   "RIGHT",     function_right },
  { FUNC_MID,     "MID",       function_mid },
  { FUNC_STR,     "STR",       function_str },
  { FUNC_CHR,     "CHR",       function_chr },
  { FUNC_DATE,    "DATE",      function_date },
  { FUNC_TIME,    "TIME",      function_time },
  { STRING_LOWER, "LOWER",     function_string_lower },
  { STRING_UPPER, "UPPER",     function_string_upper },
  { STRING_CONCAT,"CONCAT",    function_string_concat },
  { STRING_EQ,    "EQ",        function_string_eq },
  { STRING_NE,    "NE",        function_string_ne },
  { FUNC_PRINT,   "PRINT",     function_print },
  { FUNC_PRINTSPACE,   "PRINT SPACE",     function_printspace },
  { FUNC_PRINTLN,   "PRINT CRLF",     function_println },
  { 0, "NOP", 0L}
  };

/* error look-up table */
struct error {
  int code;
  char* description;
} errors[] = {
  { CXERROR_SOURCEFILE, "can't access source file" },
  { CXERROR_PARSE, "parse error" },
  { CXERROR_MEMORY, "out-of-memory" },
  { CXERROR_NESTEDSUB, "nested subroutine not allowed" },
  { CXERROR_NOBYTECODE, "program is empty" },
  { CXERROR_INVALIDBYTECODE, "invalid bytecode" },
  { CXERROR_STACK, "stack error" },
  { CXERROR_NULLREF, "null-pointer reference" },
  { CXERROR_BADSYMTYPE, "bad symbol type" },
  { CXERROR_TYPEMISMATCH, "type mismatch" },
  { CXERROR_DIVZERO, "divide by zero" },
  { CXERROR_DOMERR, "math domain error" },
  { CXERROR_NEGSQRT, "square root of negative number" },
  { CXERROR_SUB, "subroutine not found" },
  { 0, "?" }
  };


/* ========================================================================= */
/* EXPORTED FUNCTIONS not visible from cubix.h */
/* ========================================================================= */

/* create a new symbol or return the previously created symbol */
struct CXSymbol* create_symbol(char* name, int type)
{
  struct CXSymbol* ptr, *sym;
  char *str;

  /* are we in main program or in subroutine ? */
  if (current_sub) ptr = current_sub->symbols;
  else ptr = symbols;

  /* browse and look for matched name & type */
  sym = (struct CXSymbol*) 0L;
  for (; ptr; ptr = ptr->next)
    if (ptr) if (!strcasecmp(ptr->name, name))
      if (ptr->type == type) sym = ptr;

  /* if found, just return it to the caller */
  if (sym) return sym;

  /* now we must create a new one */
  str = cubix_strdup(name);
  if (!str){
    /* generate error and return NULL */
    compile_error_code = CXERROR_MEMORY;
    return (struct CXSymbol*) 0L;
    }

  /* allocate a new symbol */  
  ptr = (struct CXSymbol*) cubix_malloc (sizeof(struct CXSymbol));
  if (!ptr){
    /* generate error and return NULL */
    compile_error_code = CXERROR_MEMORY;
    cubix_free(str);
    return(struct CXSymbol*) 0L;
    };

  /* fill in value */
  ptr->name = str;
  ptr->type = type;
  ptr->external = 0;
  ptr->number = 0.0;
  ptr->string = (char*) 0L;
  ptr->next = (struct CXSymbol*) 0L;

  /* insert in front of the link-list */
  if (current_sub){
    /* local within functions */
    if (current_sub->symbols) ptr->next = current_sub->symbols;
    current_sub->symbols = ptr;
    }
  else {
    /* main program */
    if (symbols) ptr->next = symbols;
    symbols = ptr;
    }

  return ptr;

}

/* create a number */
double create_number(char* s)
{
   return atof(s);
}

/* create a new string, remove the double-quotes */
char* create_string(char *s)
{
  char *p;

  p = cubix_strdup(s);
  if (!p) return p;
  if (*p == '"') memcpy (p, p+1, strlen(p));
  if (p[strlen(p)-1] == '"') p[strlen(p)-1] = '\0';
  return p;
}

/* make a symbol EXTERNAL, i.e.: defined somewhere else */
void externalize (struct CXSymbol* symbol)
{
  if (!symbol) return;
  symbol->external = -1;
}

/* spawn for a new subroutine */
void begin_subroutine(char* name)
{
  struct CXSubroutine *sub;

  /* nested sub is not allowed */
  if (current_sub){
    compile_error_code = CXERROR_NESTEDSUB;
    return;
    }

  /* construct a new subroutine */
  sub = (struct CXSubroutine*) cubix_malloc(sizeof(struct CXSubroutine));

  /* fail to allocate space ? */
  if (!sub){
    compile_error_code = CXERROR_MEMORY;
    return;
    }

  /* copy function name */
  sub->name = cubix_strdup(name);

  /* fail to allocate space ? */
  if (!sub->name){
    compile_error_code = CXERROR_MEMORY;
    return;
    }

  /* initialize member variables */
  //printf ("subroutine [%s]\n", name);
  sub->symbols = (struct CXSymbol*) 0L;
  sub->bytecodes = (struct CXBytecode*) 0L;
  sub->next = (struct CXSubroutine*) 0L;

  /* attach to the linked list */
  if (subroutines) sub->next = subroutines;
  subroutines = sub;

  /* set also current subroutine */
  current_sub = sub;

  return;
}

/* create a bytecode to push a symbol */
void gen_pushsymbol (struct CXSymbol* symbol)
{
  struct CXBytecode *code;

  ADD_CODE(code);
  code->kind = PCODE_PUSHSYMBOL;
  code->symbol = symbol;
}

/* create a bytecode to pop stack to a symbol */
void gen_popsymbol (struct CXSymbol* symbol)
{
  struct CXBytecode *code;

  ADD_CODE(code);
  code->kind = PCODE_POPSYMBOL;
  code->symbol = symbol;
}

/* create a bytecode to push a number */
void gen_pushnumber (double number)
{
  struct CXBytecode *code;

  ADD_CODE(code);
  code->kind = PCODE_PUSHNUMBER;
  code->number = number;
}

/* create a bytecode to push a string */
void gen_pushstring (char *string)
{
  struct CXBytecode *code;

  ADD_CODE(code);
  code->kind = PCODE_PUSHSTRING;
  code->string = string;
}

/* create a conditional branch which jump to given autolabel if FALSE  */
void gen_if (void)
{
  struct CXBytecode *code;
  int i;

  i = autolabel++;
  pushint(i);
  ADD_CODE(code);
  code->kind = PCODE_IF;
  code->label = cubix_strdup(magiclabel(i));
}

/* goto, then create autolabel */
void gen_then (void)
{
  int i;

  i = autolabel++;
  pushint(i);
  gen_jump(magiclabel(i));

  swapint();
  i = popint();
  gen_label(magiclabel(i));
}

/* create terminating label  */
void gen_endif (void)
{
  int i;

  i = popint();
  gen_label(magiclabel(i));
}

/* create necessary instructions for a loop */
void gen_for (struct CXSymbol* symbol, double step)
{
  struct CXBytecode* code;
  int i;

  gen_function(SWAP_STACK);
  gen_popsymbol (symbol);

  i = autolabel++;
  pushint(i);
  gen_label(magiclabel(i));

  gen_function (COPY_STACK);
  gen_pushsymbol (symbol);

  gen_function (SWAP_STACK);

  if (step > 0) gen_function (COMPARE_LE);
  else gen_function (COMPARE_GE);

  i = autolabel++;
  pushint(i);
  ADD_CODE(code);
  code->kind = PCODE_IF;
  code->label = cubix_strdup(magiclabel(i));
}

/* loop termination */
void gen_next (struct CXSymbol* symbol, double step)
{
  int i;

  gen_pushsymbol (symbol);
  gen_pushnumber (step);
  gen_function (BINARY_ADD);
  gen_popsymbol (symbol);

  swapint();
  i = popint();
  gen_jump (magiclabel(i));
  i = popint();
  gen_label (magiclabel(i));

  gen_function (POP_STACK);
}

/* add operator or function bytecode */
void gen_function(int func)
{
  struct CXBytecode *code;

  ADD_CODE(code);
  code->kind = PCODE_FUNCTION;
  code->function = func;
}

void gen_call(char* name)
{
  struct CXBytecode *code;

  ADD_CODE(code);
  code->kind = PCODE_CALL;
  code->label = cubix_strdup(name);
}

/* end of a subroutine */
void gen_return(void)
{
  struct CXBytecode *code;

  ADD_CODE(code);
  code->kind = PCODE_RETURN;

  current_sub = (struct CXSubroutine*) 0L;
}

/* ========================================================================= */
/* EXPORTED FUNCTIONS visible from cubix.h */
/* ========================================================================= */

/* COMPATIBILITY MACRO for srandom() in glibc */

/* handle Borland C++ 5.0 */
#ifdef __BORLANDC__
#ifndef srandom
#define srandom(x) srand((x))  /* srand() is in stdlib.h */
#endif
#endif

/* handle Turbo C++ 3.0 */
#ifdef __TURBOC__
#ifndef srandom
#define srandom(x) srand((x))  /* srand() is in stdlib.h */
#endif
#endif

/* perform some initializations */
void CXInitialize (struct CXContext* cx)
{
  struct tm *timep;
  time_t atime;

  /* perform initialization of random seed */
  /* NOTE: check out the macros to handle non-POSIX system */
  time(&atime); timep=localtime(&atime);
  srandom(timep->tm_sec * timep->tm_min);

  /* initialize this context */
  cx->state = CUBIX_INITIALIZED;            /* initial state */
  cx->source = (char*) 0L;                  /* no source filename yet */
  cx->errno = 0;                            /* nothing happened yet */
  cx->execpoint = (struct CXBytecode*) 0L;  /* starting program counter is 0 */
  cx->bytecodes = (struct CXBytecode*) 0L;  /* still has no bytecode */
  cx->symbols = (struct CXSymbol*) 0L;      /* and also no symbols */
  cx->stacks = (struct CXStack*) 0L;        /* stacks aren't allocated */

  /* here we default any callbacks to nothing */
  cx->error_function = (CUBIX_CALLBACK_ERROR) 0L;
  cx->setnumber_function = (CUBIX_CALLBACK_SETNUMBER) 0L;
  cx->setstring_function = (CUBIX_CALLBACK_SETSTRING) 0L;
  cx->getnumber_function = (CUBIX_CALLBACK_GETNUMBER) 0L;
  cx->getstring_function = (CUBIX_CALLBACK_GETSTRING) 0L;
}

void CXCompile(struct CXContext* cx, char* sourcefile)
{
  extern FILE* yyin;
  extern void yyparse(void);
  extern long compile_lineno;

  /* reset local variables first */
  symbols = (struct CXSymbol*) 0L;
  bytecodes = (struct CXBytecode*) 0L;
  subroutines = (struct CXSubroutine*) 0L;

  /* setup variables */
  cx->state = CUBIX_COMPILE;
  autolabel = 0;
  compile_lineno = 1;
  compile_error_code = 0;
  intstack_count = 0;
  current_sub = (struct CXSubroutine*) 0L;

  /* load the sourcefile */
  if (sourcefile) yyin = fopen (sourcefile, "rt");
  else yyin = stdin;

  /* if failed, trigger an error and return */
  if (!yyin){
     trigger_error(cx, CXERROR_SOURCEFILE);
     return;
     }

  /* scan/parse/generate bytecodes */
  yyparse();

  /* close source file */
  fclose(yyin);

  /* check for error ? */
  if (compile_error_code){
    cx->lineno = compile_lineno;
    trigger_error(cx, compile_error_code);
    return;
    }

  /* make sure there's PCODE_TERMINATE */
  gen_terminate();

  /* copy to the context */
  cx->symbols = symbols;
  cx->bytecodes = bytecodes;
  cx->subroutines = subroutines;

  /* ready to be executed */
  cx->state = CUBIX_EXECUTE;
}

/* run the context */
void CXExecute(struct CXContext* cx)
{
  void explain (struct CXContext*);

  /* sentinel only! */
  if (!cx) return;

  /* don't run upon error */
  if (cx->state == CUBIX_ERROR) return;

  /* release any previous stack */
  cx->stacks = release_stacks(cx->stacks);

  /* set the context */
  cx->state = CUBIX_EXECUTE;
  cx->execpoint = cx->bytecodes;

  while (cx->execpoint){
   //explain(cx); cx->execpoint = cx->execpoint->next;
   single_step(cx);
   }
}

/* run a subroutine */
void CXExecuteSub(struct CXContext* cx, char* name)
{
  /* sentinel only! */
  if (!cx) return;

  /* don't run upon error */
  if (cx->state == CUBIX_ERROR) return;

  /* look for subroutine's name */
  cx->state = CUBIX_EXECUTE;
  program_call(cx, name);

  /* give NULL return address first */
  STACK_PUSHADDRESS(cx, (struct CXBytecode*) 0L);

  while (cx->execpoint){
   //explain(cx); cx->execpoint = cx->execpoint->next;
   single_step(cx);
   }
}


/* release the stack */
struct CXStack* release_stacks(struct CXStack* stack)
{
  struct CXStack* nextstack;

  /* sentinel only */
  if (!stack) return stack;

  /* release the stack */
  for (; stack;){
    nextstack = stack->next;
    stack->string = cubix_free(stack->string); /* free associate string */
    cubix_free(stack);
    stack = nextstack;
    }

  /* always return NULL */
  return (struct CXStack*) 0L;

}

/* release bytecode */
struct CXBytecode* release_bytecodes(struct CXBytecode* code)
{
  struct CXBytecode* nextcode;

  /* sentinel only */
  if (!code) return code;

  for (; code;){
    nextcode = code->next;
    code->string = cubix_free(code->string); /* immediate value must be freed */
    code->label = cubix_free(code->label);
    cubix_free(code);
    code = nextcode;
    }

  /* always return NULL */
  return (struct CXBytecode*) 0L;


}

/* release symbols */
struct CXSymbol* release_symbols(struct CXSymbol* symbol)
{
  struct CXSymbol* nextsymbol;

  /* sentinel only ! */
  if (!symbol) return symbol;

  /* release all symbols */
  for (; symbol; ){
    nextsymbol = symbol->next;
    symbol->name = cubix_free(symbol->name);
    symbol->string = cubix_free(symbol->string); /* free string first */
    cubix_free(symbol);
    symbol = nextsymbol;
    }

  /* always return NULL */
  return (struct CXSymbol*) 0L;

}

/* release dynamically alloc'ed space in the context */
void CXRelease(struct CXContext* cx)
{
  struct CXSubroutine* sub, *nextsub;

  /* sentinel only */
  if (!cx) return;

  /* free our stack first */
  cx->stacks = release_stacks(cx->stacks);

  /* release stuff of main program */
  cx->bytecodes = release_bytecodes(cx->bytecodes);
  cx->symbols = release_symbols(cx->symbols);

  /* for each function, release local stuff */
  for (sub = cx->subroutines; sub; ){
    nextsub = sub->next;
    sub->name = cubix_free(sub->name);
    sub->symbols = release_symbols(sub->symbols);
    sub->bytecodes = release_bytecodes(sub->bytecodes);
    cubix_free(sub);
    sub = nextsub;
    }
}

/* return description of error number */
char* CXDescribeError (struct CXContext* cx)
{
  int i, err;

  /* get error code, search in look-up table */
  err = cx->errno;
  for (i = 0; errors[i].code; i++)
    if (errors[i].code == err)
      return errors[i].description;

  /* unknown error code ? PLEASE check the lookup table */
  return "Unknown";
}


/* ========================================================================= */
/*  MEMORY MANAGEMENT FUNCTIONS  */
/* ========================================================================= */

void* cubix_malloc(size_t size)
{
  cubix_malloc_count++;
  return malloc(size);
}

void* cubix_free(void* ptr)
{
  if (!ptr) return ptr;
  free(ptr);
  cubix_malloc_count--;
  return (void*) 0L;
}

char* cubix_strdup(char* string)
{
  cubix_malloc_count++;
  return strdup(string);
}

/* ========================================================================= */
/*  LOCAL FUNCTIONS (for the compiler) */
/* ========================================================================= */

/* push an integer to stack */
void pushint(int x)
{
  intstacks[intstack_count] = x;
  intstack_count++;
}

/* pop an integer from stack */
int popint(void)
{
  intstack_count--;
  return intstacks[intstack_count];
}

/* swap two topmost integers on stack */
void swapint(void)
{
  int temp;

  temp = intstacks[intstack_count-1];
  intstacks[intstack_count-1] = intstacks[intstack_count-2];
  intstacks[intstack_count-2] = temp;
}

/* compiler error, to be called from yyparse() */
/* TO DO: create a verbose error handler */
void compile_error (long lineno, char* msg, char* yytext)
{
  compile_error_code = CXERROR_PARSE;
}

/* create a bytecode to terminate program */
void gen_terminate (void)
{
  struct CXBytecode *code;

  ADD_CODE(code);
  code->kind = PCODE_TERMINATE;
}

/* create a quick-jump :) */
void gen_jump (char *label)
{
  struct CXBytecode *code;

  ADD_CODE(code);
  code->kind = PCODE_JUMP;
  code->label = cubix_strdup(label);
}

/* create a label */
void gen_label (char *label)
{
  struct CXBytecode *code;

  ADD_CODE(code);
  code->kind = PCODE_LABEL;
  code->label = cubix_strdup(label);
}

/* add a new bytecode, return pointer to the code (NULL upon error) */
struct CXBytecode* code_add(void)
{
  struct CXBytecode *code, *ptr;

  /* construct a new code */
  code = (struct CXBytecode*) cubix_malloc(sizeof(struct CXBytecode));

  /* unable to allocate space ? just quit then.. */
  if (!code){
    compile_error_code = CXERROR_MEMORY;
    return code;
    }

  /* initialize default variables */
  code->kind = -1;
  code->function = 0;
  code->number = 0.0;
  code->symbol = (struct CXSymbol*) 0L;
  code->label = (char*) 0L;
  code->string = (char*) 0L;
  code->next = (struct CXBytecode*) 0L;

  /* main program or in subroutine ? */
  if (!current_sub){

     /* attach to end-of-list */
     if (!bytecodes) bytecodes = code;
     else {
        for (ptr = bytecodes; ptr->next; ptr = ptr->next);
        ptr->next = code;
       }
     }
  else {

     /* attach to function bytecodes */
     if (!current_sub->bytecodes) current_sub->bytecodes = code;
     else {
        for (ptr = current_sub->bytecodes; ptr->next; ptr = ptr->next);
        ptr->next = code;
       }
    }

  return code;
}

/* ========================================================================= */
/*  LOCAL FUNCTIONS (for the virtual-machine) */
/* ========================================================================= */

/* trigger an error */
void trigger_error(struct CXContext *cx, int code)
{
  char *desc;

  /* sentinel only ! */
  if (!cx) return;

  /* set context state */
  cx->state = CUBIX_ERROR;

  /* store error number */
  cx->errno = code;

  /* call the callback error function */
  desc = CXDescribeError(cx);
  if (cx->error_function) (*cx->error_function)(desc);
  else {
    /* no error callback functions, so here is our own version */
    printf ("Error: %s\n", desc);
    exit(cx->errno);
    }
}

/* return pointer to subroutine's bytecode list */
struct CXSubroutine* get_subroutine(struct CXContext* cx, char* name)
{
  struct CXSubroutine* sub;
  struct CXSubroutine* result;

  /* sentinel only ! */
  if (!cx) return (struct CXSubroutine*) 0L;

  /* check for error first */
  if (cx->state == CUBIX_ERROR) return (struct CXSubroutine*) 0L;

  /* look for subroutine's name */
  result = (struct CXSubroutine*) 0L;
  for (sub = cx->subroutines; sub; sub = sub->next)
    if (!strcasecmp(sub->name, name))
      result = sub;

  return result;
}

/* run one bytecode only :) */
void single_step(struct CXContext* cx)
{
  struct CXBytecode *code;
  struct CXSymbol* symbol;
  struct CXStack* st;
  double x; long i;

  /* must be in CUBIX_EXECUTE */
  if (cx->state != CUBIX_EXECUTE) return;

  /* sentinel */
  code = cx->execpoint;
  if (!code){

     /* generate error: execution point is NULL ! */
     trigger_error (cx, CXERROR_NOBYTECODE);
     return;
     }

  switch (code->kind){

    /* mark end-of-execution, terminate the context */
    case PCODE_TERMINATE:
      cx->state = CUBIX_TERMINATED;
      cx->execpoint = (struct CXBytecode*) 0L;
      break;

    /* call a subroutine */
    case PCODE_CALL:
      STACK_PUSHADDRESS(cx, cx->execpoint->next);
      program_call (cx, code->label);
      break;

    /* return from a subroutine */
    case PCODE_RETURN:

      /* assume return address isn't provided */
      cx->execpoint = (struct CXBytecode*) 0L;

      /* what is topmost entry in stack ? */
      st = cx->stacks;
      if (!st){
      trigger_error(cx, CXERROR_NULLREF);
      return;
       }

      /* return to caller */
      if (st->type == CUBIX_ADDRESS)
        STACK_POPADDRESS(cx, cx->execpoint);

      break;

    /* push a symbol, check symbol type first */
    case PCODE_PUSHSYMBOL:
      symbol = code->symbol;
      switch (symbol->type){
      case CUBIX_NUMBER:
        if ((symbol->external) && (cx->getnumber_function))
          symbol->number = cx->getnumber_function(symbol->name);
        STACK_PUSHNUMBER(cx, symbol->number);
        break;
      case CUBIX_STRING:
        if ((symbol->external) && (cx->getstring_function))
          symbol->string = cx->getstring_function(symbol->name);
        STACK_PUSHSTRING(cx, symbol->string);
        break;
      default:
        trigger_error (cx, CXERROR_BADSYMTYPE);
        return;
      }
      cx->execpoint = cx->execpoint->next;
      break;

    /* pop a symbol, also must check symbol type and external flag */
    case PCODE_POPSYMBOL:
      symbol = code->symbol;
      switch(symbol->type){
      case CUBIX_NUMBER:
        STACK_POPNUMBER(cx, symbol->number);
        if ((symbol->external) && (cx->setnumber_function))
          cx->setnumber_function (symbol->name, symbol->number);
        break;
      case CUBIX_STRING:
        cubix_free(symbol->string);
        STACK_POPSTRING(cx, symbol->string);
        if ((symbol->external) && (cx->setstring_function))
          cx->setstring_function (symbol->name, symbol->string);
        break;
      default:
        trigger_error (cx, CXERROR_BADSYMTYPE);
        return;
      }
      cx->execpoint = cx->execpoint->next;
      break;

    /* push a number to stack */
    case PCODE_PUSHNUMBER:
      STACK_PUSHNUMBER(cx, code->number);
      cx->execpoint = cx->execpoint->next;
      break;

    /* push a string to stack */
    case PCODE_PUSHSTRING:
      STACK_PUSHSTRING(cx, code->string);
      cx->execpoint = cx->execpoint->next;
      break;

    /* jump to a given label */
    case PCODE_JUMP:
      program_jump (cx, code->label);
      break;

    /* if popped number false then jump to a given label, else continue */
    case PCODE_IF:
      STACK_POPNUMBER(cx, x);
      if (MAKE_LOGICAL(x)) cx->execpoint = cx->execpoint->next;
      else program_jump (cx, code->label);
      break;

    /* search function in callback-table and call apropriate function */
    case PCODE_FUNCTION:
      for (i = 0; functions[i].type; i++)
        if (code->function == functions[i].type)
          if (functions[i].proc) (*functions[i].proc)(cx);
      cx->execpoint = cx->execpoint->next;
      break;

    /* a label, just ignore it */
    case PCODE_LABEL:
      cx->execpoint = cx->execpoint->next;
      break;

    /* this should be a really bad bytecode */
    default:
      trigger_error(cx, CXERROR_INVALIDBYTECODE);
      cx->execpoint = cx->execpoint->next;
      break;
   }
}


/* push a number to stack */
#define RT_ASSERT(ptr,ret) if (!ptr) { /* generate error */; return (ret); }
struct CXStack* stack_pushnumber (struct CXContext* cx, double number)
{
  struct CXStack* st;

  /* allocate new entry */
  st = (struct CXStack*) cubix_malloc(sizeof(struct CXStack));

  /* if allocation is failed, trigger error */
  if (!st){
    /* trigger error */
    return st;
    }

  /* fill in value for a number */
  st->type = CUBIX_NUMBER;
  st->number = number;
  st->string = (char*) 0L;
  st->code = (struct CXBytecode*) 0L;

  /* link to the stack list */
  st->next = (struct CXStack*) 0L;
  if (cx->stacks) st->next = cx->stacks;
  cx->stacks = st;
  return st;
}

/* push a string to stack, string is duplicated */
struct CXStack* stack_pushstring (struct CXContext* cx, char *string)
{
  struct CXStack* st;

  st = (struct CXStack*) cubix_malloc(sizeof(struct CXStack));
  RT_ASSERT(st,st);
  st->type = CUBIX_STRING;
  st->number = 0.0;
  st->string = cubix_strdup(string);
  st->code = (struct CXBytecode*) 0L;
  RT_ASSERT(st->string, (struct CXStack*) 0L);
  st->next = (struct CXStack*) 0L;
  if (cx->stacks) st->next = cx->stacks;
  cx->stacks = st;
  return st;
}

/* push a pointer to bytecode */
struct CXStack* stack_pushaddress(struct CXContext* cx, struct CXBytecode* code)
{
  struct CXStack* st;

  /* allocate new entry */
  st = (struct CXStack*) cubix_malloc(sizeof(struct CXStack));

  /* if allocation is failed, trigger error */
  if (!st){
    /* trigger error */
    trigger_error(cx, CXERROR_MEMORY);
    return st;
    }

  /* fill in value for a number */
  st->type = CUBIX_ADDRESS;
  st->number = 0.0;
  st->string = (char*) 0L;
  st->code = code;

  /* link to the stack list */
  st->next = (struct CXStack*) 0L;
  if (cx->stacks) st->next = cx->stacks;
  cx->stacks = st;
  return st;
}

/* pop a number from stack */
double stack_popnumber (struct CXContext* cx)
{
  struct CXStack* st;
  double number;

  st = cx->stacks;
  RT_ASSERT(st, (double)0L);
  CHECK_TYPE(cx, st->type, CUBIX_NUMBER);
  number = st->number;
  cx->stacks = cx->stacks->next;
  cubix_free(st);
  return number;
}

/* pop a string from stack */
char* stack_popstring (struct CXContext* cx)
{
  struct CXStack* st;
  char *string;

  st = cx->stacks;
  RT_ASSERT(st, (char*)0L);
  CHECK_TYPE(cx, st->type, CUBIX_STRING);
  string = st->string;
  cx->stacks = cx->stacks->next;
  cubix_free(st);
  return string;
}

/* pop an address from stack */
struct CXBytecode* stack_popaddress(struct CXContext* cx)
{
  struct CXStack* st;
  struct CXBytecode* code;

  st = cx->stacks;
  if (!st){
    trigger_error(cx, CXERROR_STACK);
    return (struct CXBytecode*) 0L;
    }

  CHECK_TYPE(cx, st->type, CUBIX_ADDRESS);
  code = st->code;
  cx->stacks = cx->stacks->next;
  cubix_free(st);
  return code;
}



/* swap topmost entries in stack */
void stack_swap (struct CXContext* cx)
{
  struct CXStack *p1, *p2;

  if (!cx->stacks) return;
  if (!cx->stacks->next) return;
  p1 = cx->stacks;
  p2 = cx->stacks->next;
  p1->next = p2->next;
  p2->next = p1;
  cx->stacks = p2;
}

/* copy topmost entry on stack */
void stack_copy (struct CXContext* cx)
{
  struct CXStack* st;

  st = cx->stacks;
  if (!st) return;
  if (st->type == CUBIX_NUMBER) stack_pushnumber(cx, st->number);
  if (st->type == CUBIX_STRING) stack_pushstring(cx, st->string);
}

/* pop entry from stack */
void stack_pop (struct CXContext* cx)
{
  struct CXStack* st;

  st = cx->stacks;
  if (!st) return;
  cx->stacks = cx->stacks->next;
  cubix_free(st);
}

/* return description of given function code */
char* describe_function (int func)
{
  long i;
  for (i = 0; functions[i].type; i++)
   if (func == functions[i].type)
      return functions[i].name;
  return "<unknown>";
}

/* jump to a label */
void program_jump(struct CXContext* cx, char* label)
{
  struct CXBytecode *code, *jump;
  struct CXSubroutine *sub;

  /* sentinel */
  if (!cx->bytecodes) return;

  /* search in main program */
  jump = (struct CXBytecode*) 0L;
  for (code = cx->bytecodes; code; code = code->next)
    if (code->kind == PCODE_LABEL) if (code->label)
      if (!strcasecmp(code->label, label))
        jump = code;

  /* search in each subroutine */
  for (sub = cx->subroutines; sub; sub = sub->next){
    for (code = sub->bytecodes; code; code = code->next)
      if (code->kind == PCODE_LABEL) if (code->label)
        if (!strcasecmp(code->label, label))
          jump = code;
    }

  if (jump) cx->execpoint = jump;
  else if (cx->execpoint) cx->execpoint = cx->execpoint->next;
}

/* call a function */
void program_call(struct CXContext* cx, char* name)
{
  struct CXSubroutine* jump;

  /* sentinel only! */
  if (!cx) return;

  /* don't run upon error */
  if (cx->state == CUBIX_ERROR) return;

  /* look for subroutine */
  jump = get_subroutine(cx, name);

  /* jump to given function */
  if (jump)
    cx->execpoint = jump->bytecodes;
  else {
    /* function not found */
    trigger_error(cx, CXERROR_SUB);
    cx->execpoint = cx->execpoint->next;
    }
}

/* negate a number */
void function_negate (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  STACK_PUSHNUMBER(cx, -x);
}

/* add two numbers */
void function_add (struct CXContext* cx)
{
  double x, y;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  STACK_PUSHNUMBER(cx, x + y);
}

/* substract two numbers */
void function_sub (struct CXContext* cx)
{
  double x, y;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  STACK_PUSHNUMBER(cx, x - y);
}

/* multiply two numbers */
void function_mul (struct CXContext* cx)
{
  double x, y;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  STACK_PUSHNUMBER(cx, x * y);
}

/* divide two numbers */
void function_div (struct CXContext* cx)
{
  double x, y;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  if (IS_SMALL(y)) trigger_error (cx, CXERROR_DIVZERO);
  else STACK_PUSHNUMBER(cx, x / y);
}

/* calculate x^y */
void function_power (struct CXContext* cx)
{
  double x, y;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  if (y < 0) trigger_error (cx, CXERROR_DOMERR);
  else STACK_PUSHNUMBER(cx, exp(y*log(x)));
}

/* bitwise not operation */
void function_not (struct CXContext* cx)
{
  double x;
  long c;

  STACK_POPNUMBER(cx,x);
  c = (long) x; c = ~c; x = (double) c;
  STACK_PUSHNUMBER(cx,x);
}

/* bitwise or operation */
void function_or (struct CXContext* cx)
{
  double x, y, z;
  long a, b, c;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  a = (long) x;
  b = (long) y;
  c = a | b;
  z = (double) c;
  STACK_PUSHNUMBER(cx, z);
}

/* bitwise and operation */
void function_and (struct CXContext* cx)
{
  double x, y, z;
  long a, b, c;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  a = (long) x;
  b = (long) y;
  c = a & b;
  z = (double) c;
  STACK_PUSHNUMBER(cx, z);
}

/* bitwise xor operation */
void function_xor (struct CXContext* cx)
{
  double x, y, z;
  long a, b, c;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  a = (long) x;
  b = (long) y;
  c = a ^ b;
  z = (double) c;
  STACK_PUSHNUMBER(cx, z);
}

/* logical not operation */
void function_boole_not (struct CXContext* cx)
{
  double x;
  long c;

  STACK_POPNUMBER(cx,x);
  c = MAKE_LOGICAL(x); c = !c; x = (double) c;
  STACK_PUSHNUMBER(cx,x);
}

/* logical or operation */
void function_boole_or (struct CXContext* cx)
{
  double x, y, z;
  long a, b, c;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  a = MAKE_LOGICAL(x);
  b = MAKE_LOGICAL(y);
  c = a || b;
  z = (double) c;
  STACK_PUSHNUMBER(cx, z);
}

/* logical and operation */
void function_boole_and (struct CXContext* cx)
{
  double x, y, z;
  long a, b, c;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  a = MAKE_LOGICAL(x);
  b = MAKE_LOGICAL(y);
  c = a && b;
  z = (double) c;
  STACK_PUSHNUMBER(cx, z);
}

/* true if numbers are equal */
void function_compare_eq (struct CXContext* cx)
{
  double x, y, z;
  long c;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  c = (x == y); c = MAKE_LOGICAL(c);
  z = (double) c;
  STACK_PUSHNUMBER(cx, z);
}

/* false if numbers are equal */
void function_compare_ne (struct CXContext* cx)
{
  double x, y, z;
  long c;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  c = (x != y); c = MAKE_LOGICAL(c);
  z = (double) c;
  STACK_PUSHNUMBER(cx, z);
}

/* true if first number less than second number */
void function_compare_lt (struct CXContext* cx)
{
  double x, y, z;
  long c;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  c = (x < y); c = MAKE_LOGICAL(c);
  z = (double) c;
  STACK_PUSHNUMBER(cx, z);
}

/* true if first number less than or equal second number */
void function_compare_le (struct CXContext* cx)
{
  double x, y, z;
  long c;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  c = (x <= y); c = MAKE_LOGICAL(c);
  z = (double) c;
  STACK_PUSHNUMBER(cx, z);
}

/* true if first number greater than second number */
void function_compare_gt (struct CXContext* cx)
{
  double x, y, z;
  long c;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  c = (x > y); c = MAKE_LOGICAL(c);
  z = (double) c;
  STACK_PUSHNUMBER(cx, z);
}

/* true if first number greater than or equal second number */
void function_compare_ge (struct CXContext* cx)
{
  double x, y, z;
  long c;

  STACK_POPNUMBER(cx, y);
  STACK_POPNUMBER(cx, x);
  c = (x >= y); c = MAKE_LOGICAL(c);
  z = (double) c;
  STACK_PUSHNUMBER(cx, z);
}

/* sine function, argument in radian */
void function_sin (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  x = sin(x);
  STACK_PUSHNUMBER(cx, x);
}

void function_asin (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  x = asin(x);
  STACK_PUSHNUMBER(cx, x);
}

void function_cos (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  x = cos(x);
  STACK_PUSHNUMBER(cx, x);
}

void function_acos (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  x = acos(x);
  STACK_PUSHNUMBER(cx, x);
}

void function_tan (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  x = tan(x);
  STACK_PUSHNUMBER(cx, x);
}

void function_atan (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  x = atan(x);
  STACK_PUSHNUMBER(cx, x);
}

void function_exp (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  x = exp(x);
  STACK_PUSHNUMBER(cx, x);
}

void function_log (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  x = log(x);
  STACK_PUSHNUMBER(cx, x);
}

void function_sqrt (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  if (x < 0) trigger_error (cx, CXERROR_NEGSQRT);
  else STACK_PUSHNUMBER(cx, sqrt(x));
}

void function_int (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  STACK_PUSHNUMBER(cx, floor(x));
}

void function_frac (struct CXContext* cx)
{
  double x;

  STACK_POPNUMBER(cx, x);
  STACK_PUSHNUMBER(cx, x - floor(x));
}

/* return random number in 0..1 */
void function_ran (struct CXContext* cx)
{
  double x;
  x = (double)rand() / RAND_MAX;
  STACK_PUSHNUMBER(cx, x);
}

/* get length of string */
void function_len (struct CXContext* cx)
{
  char *string;
  double x;

  STACK_POPSTRING(cx, string);
  x = (double) strlen(string);
  STACK_PUSHNUMBER(cx, x);
  cubix_free(string);  /* allocated memory for the string has to be freed */
}

/* convert a string to a number */
void function_val (struct CXContext* cx)
{
  char *string;
  double x;
  int i;

  STACK_POPSTRING(cx, string);
  i = sscanf (string, "%lf", &x);
  if (i != 1) x = 0;
  STACK_PUSHNUMBER(cx, x);
  cubix_free(string); /* allocated memory for the string has to be freed */
}

/* get ASCII value of first char in string */
void function_asc (struct CXContext* cx)
{
  char *string;
  double x;

  STACK_POPSTRING(cx, string);
  x = (double) string[0];
  STACK_PUSHNUMBER(cx, x);
  cubix_free(string); /* allocated memory for the string has to be freed */
}

/* get first n characters from a string */
void function_left (struct CXContext* cx)
{
  char *string;
  double x;
  int i;

  STACK_POPNUMBER(cx, x);
  if (x < 0) x = 0; i = (int) x;
  STACK_POPSTRING(cx, string);

  /* truncate the string */
  if (i < strlen(string)) string[i] = '\0';

  /* push the string, it's safe cause pushstring will do strdup() */
  STACK_PUSHSTRING(cx, string);

  /* free previous string */
  cubix_free(string);
}

/* get last n characters from a string */
void function_right (struct CXContext* cx)
{
  char *string;
  double x;
  int i;

  STACK_POPNUMBER(cx, x);
  if (x < 0) x = 0; i = (int) x;
  STACK_POPSTRING(cx, string);

  /* copy n last characters  */
  if (i < strlen(string))
    memcpy (string, string + strlen(string) - i, i + 1);

  /* push the string, it's safe cause pushstring will do strdup() */
  STACK_PUSHSTRING(cx, string);

  /* free previous string */
  cubix_free(string);
}

/* grab portion of a string */
void function_mid (struct CXContext* cx)
{
  char *string;
  double x;
  int i, j;

  STACK_POPNUMBER(cx, x);
  if (x < 0) x = 0; j = (int) x;
  STACK_POPNUMBER(cx, x);
  if (x < 0) x = 0; i = (int) x;
  STACK_POPSTRING(cx, string);

  /* more than length of string ? */
  if (i > strlen(string)) i = strlen(string);
  if (i + j > strlen(string)) j = strlen(string) - i;

  /* copy */
  memcpy (string, string + i - 1, j);
  string[j] = '\0';

  /* push the string, it's safe cause pushstring will do strdup() */
  STACK_PUSHSTRING(cx, string);

  /* free previous string */
  cubix_free(string);
}

/* convert number to string representation */
#define MAX_STR_RESULT 64
void function_str (struct CXContext* cx)
{
  static char result[MAX_STR_RESULT];
  double x;

  STACK_POPNUMBER(cx, x);
  sprintf (result, "%g", x);
  STACK_PUSHSTRING(cx, result);
}

/* get character of given ASCII code */
void function_chr (struct CXContext* cx)
{
  double x;
  static char result[2];

  STACK_POPNUMBER(cx, x);
  result[0] = (char) x;
  result[1] = '\0';
  STACK_PUSHSTRING(cx, result);
}

/* return current date, format is dd.mm.yyyy with fixed length of 10 chars */
void function_date (struct CXContext* cx)
{
  struct tm *timep;
  time_t atime;
  char result[11];

  /* get current date & time */
  time(&atime); timep=localtime(&atime);

  /* fill the result with current date*/
  sprintf (result, "%02i.%02i.%04i",
    timep->tm_mday, timep->tm_mon+1, timep->tm_year+1900);

  /* push to stack */
  STACK_PUSHSTRING(cx, result);
}

/* return current time, format is hh:mm:ss with fixed length of 8 chars */
void function_time (struct CXContext* cx)
{
  struct tm *timep;
  time_t atime;
  char result[9];

  /* get current date & time */
  time(&atime); timep=localtime(&atime);

  /* fill the result with current time*/
  sprintf (result, "%02i:%02i:%02i",
      timep->tm_hour, timep->tm_min, timep->tm_sec);

  /* push to stack */
  STACK_PUSHSTRING(cx, result);
}

/* convert uppercase letters in a string to lowercase */
void function_string_lower(struct CXContext* cx)
{
  char *result;
  int i;

  /* get the string */
  STACK_POPSTRING(cx, result);

  /* convert to lowercase */
  for(i = 0; i < strlen(result); i++)
    result[i] = tolower(result[i]);

  /* push the result */
  STACK_PUSHSTRING(cx, result);

  /* must be freed since result is strdup()'ed */
  cubix_free(result);
}

/* convert lowercase letters in a string to uppercase */
void function_string_upper(struct CXContext* cx)
{
  char *result;
  int i;

  /* get the string */
  STACK_POPSTRING(cx, result);

  /* convert to lowercase */
  for(i = 0; i < strlen(result); i++)
    result[i] = toupper(result[i]);

  /* push the result */
  STACK_PUSHSTRING(cx, result);

  /* must be freed since result is strdup()'ed */
  cubix_free(result);
}

/* concat two strings, for the '+' string operation */
void function_string_concat(struct CXContext* cx)
{
  char *p, *q;
  char* result;

  /* get the two strings */
  STACK_POPSTRING(cx, q);
  STACK_POPSTRING(cx, p);

  /* allocate space for result */
  result = (char*) cubix_malloc(strlen(p)+strlen(q)+1);

  /* cannot malloc() ! */
  if (!result){
    trigger_error(cx, CXERROR_MEMORY);
    return;
    }

  /* concat the strings */
  strcpy(result, p);
  strcat(result, q);

  /* free the strings, not used anymore */
  cubix_free(p);
  cubix_free(q);

  /* push to stack */
  STACK_PUSHSTRING(cx, result);

  /* must be freed since result is strdup()'ed */
  cubix_free(result);
}

/* return true if two strings are equal */
void function_string_eq (struct CXContext* cx)
{
  char *p, *q;
  long c;

  /* get the two strings */
  STACK_POPSTRING(cx, q);
  STACK_POPSTRING(cx, p);

  /* compare the string, push the result */
  c = strcmp(p, q) == 0;
  c = MAKE_LOGICAL(c);
  STACK_PUSHNUMBER(cx, c);

  /* free the strings, not used anymore */
  cubix_free(p);
  cubix_free(q);
}

/* return false if two strings are equal */
void function_string_ne (struct CXContext* cx)
{
  char *p, *q;
  long c;

  /* get the two strings */
  STACK_POPSTRING(cx, q);
  STACK_POPSTRING(cx, p);

  /* compare the string, push the result */
  c = strcmp(p, q) == 0;
  c = MAKE_LOGICAL(c);
  STACK_PUSHNUMBER(cx, c);

  /* free the strings, not used anymore */
  cubix_free(p);
  cubix_free(q);
}

/* print topmost entry on stack */
void function_print (struct CXContext* cx)
{
  struct CXStack *st;
  double number;
  char *string;

  /* we have to know what is the topmost entry on stack */
  st = cx->stacks;

  /* check for stupid stack error, this is a sentinel */
  if (!st){
    trigger_error(cx, CXERROR_NULLREF);
    return;
    }

  switch(st->type){

  /* print a number */
  case CUBIX_NUMBER:
    STACK_POPNUMBER(cx, number);
    printf ("%g", number);
    break;

  /* print a string */
  case CUBIX_STRING:
    STACK_POPSTRING(cx, string);
    printf ("%s", string);
    cubix_free(string);
    break;

  /* bad type ? */
  default:
    trigger_error(cx, CXERROR_TYPEMISMATCH);
    return;
  }
}

void function_printspace(struct CXContext* cx)
{
  printf(" ");
}

void function_println(struct CXContext* cx)
{
  printf("\n");
}

/* DEBUG: explain current bytecode */
void explain (struct CXContext* cx)
{
  struct CXBytecode* code;

  code = cx->execpoint;
  if (!code) return;

  switch (code->kind){
  case PCODE_TERMINATE:
    printf ("  END\n"); break;
  case PCODE_PUSHSYMBOL:
    printf ("  PUSH   #%s\n", code->symbol->name); break;
  case PCODE_POPSYMBOL:
    printf ("  POP    #%s\n", code->symbol->name); break;
  case PCODE_PUSHNUMBER:
    printf ("  PUSH  %g\n", code->number); break;
  case PCODE_PUSHSTRING:
    printf ("  PUSH  [%s]\n", code->string); break;
  case PCODE_JUMP:
    printf ("  JUMP   %s\n", code->label); break;
  case PCODE_IF:
    printf ("  JFALSE %s\n", code->label); break;
  case PCODE_LABEL:
    printf ("%s:\n", code->label); break;
  case PCODE_FUNCTION:
    printf ("  %s\n", describe_function (code->function));  break;
  case PCODE_CALL:
    printf ("  CALL %s\n", code->label); break;
  case PCODE_RETURN:
    printf ("  RETURN\n"); break;
  default: printf ("bad pcode [%d]!\n", code->kind); break;
  }
}
